package consume

import (
	"encoding/json"
	"query-service/internal/entity"
	"query-service/pkg/errorhandler"
	"strings"

	"git.science.uu.nl/datastrophe/broker"
)

// HandleMessage gets called when a message is received
func (s *Service) HandleMessage(msg *broker.Message) {
	// Grab sessionID and clientID from the headers
	sessionID, ok := msg.Headers["sessionID"].(string)
	if !ok {
		return
	}

	clientID, ok := msg.Headers["clientID"].(string)
	if !ok {
		return
	}

	// Marshall the incoming message into an IncomingJSONQuery object
	var JSONQuery entity.IncomingQueryJSON
	err := json.Unmarshal(msg.Body, &JSONQuery)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_translation_error",
			Value: "failed to unmarshall the incoming JSON query",
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	// Convert the json byte msg to a query string
	query, databaseName, err := s.queryConverter.ConvertQuery(&msg.Body)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_translation_error",
			Value: err.Error(),
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	// Check if there was a database name
	if *databaseName == "" {
		errorMsg := entity.MessageStruct{
			Type:  "query_malformed_request_error",
			Value: "no database name supplied",
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	// TODO: Figure out a way to do this in the integration tests instead
	// Remove newlines and tabs from string
	queryNoNewlines := strings.ReplaceAll(*query, "\n", "")
	queryNoTabs := strings.ReplaceAll(queryNoNewlines, "\t", "")

	// Send the resulting query back to the client
	translationMsg := entity.MessageStruct{
		Type:  "query_translation_result",
		Value: queryNoTabs,
	}
	translationMsgBytes, err := json.Marshal(translationMsg)
	if err != nil {
		errorhandler.LogError(err, "Marshalling query_translation_result went wrong!") // TODO: send error message to client instead
		return
	}
	s.producer.PublishMessage(&translationMsgBytes, &sessionID) // TODO: should this be a go routine?

	// Fetch database credentials from the user service
	databaseInfo, err := s.databaseInfoService.GetDatabaseInfo(&clientID, databaseName)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_database_not_found",
			Value: err.Error(),
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	// execute and retrieve result
	// convert result to general (node-link) format
	result, err := s.requestSender.SendAQLQuery(*query, databaseInfo.Username, databaseInfo.Password, databaseInfo.URL, databaseInfo.Port, databaseInfo.InternalDatabaseName)
	if err != nil {
		errorMsg := entity.MessageStruct{
			Type:  "query_database_error",
			Value: err.Error(),
		}
		errorMsgBytes, _ := json.Marshal(errorMsg)
		s.producer.PublishMessage(&errorMsgBytes, &sessionID)
		return
	}

	// Add type indicator to result from database
	// TODO: Change key 'values' to key 'value'
	var res interface{}
	json.Unmarshal(*result, &res)
	resultMsg := entity.MessageStruct{
		Type:  "query_result",
		Value: res,
	}
	resultMsgBytes, err := json.Marshal(resultMsg)
	if err != nil {
		errorhandler.LogError(err, "Marshalling query_result went wrong!") // TODO: send error message to client instead
		return
	}

	s.producer.PublishMessage(&resultMsgBytes, &sessionID)
}
